#!/usr/bin/env python3
import yaml
import importlib.util
import sys
import os
import subprocess
import time
import re
from motoko.task_manager import TaskManager
from motoko.bd_study import create_bd_studies
from BlackDynamite import _transaction

################################################################


def no_action(*args, **kwargs):
    pass


################################################################


class VarsDBManager:
    """Manages the database behind Vars"""

    def __init__(self, db_fname):
        self.db_fname = db_fname
        self.db = None
        self.conn = None

    def __enter__(self):
        from ZODB import FileStorage, DB

        storage = FileStorage.FileStorage(self.db_fname)
        self.db = DB(storage)
        self.conn = self.db.open()
        return self.conn.root()

    def __exit__(self, exc_type, exc_value, tb):
        import transaction

        transaction.commit()
        self.conn.close()
        self.db.close()


################################################################


class WorkflowVars:
    """Driver class for storing and accessing workflow variables,
    e.g., when performing an action upon finishing a task
    """

    def __init__(self, db_fname):
        self.db_fname = db_fname

    def __setattr__(self, name, value):
        if name != "db_fname":
            with VarsDBManager(self.db_fname) as data:
                data[name] = value
                import transaction

                transaction.commit()

        else:
            super().__setattr__(name, value)

    def __getattr__(self, name):
        with VarsDBManager(self.db_fname) as data:
            if name in data:
                return data[name]

        return super().__getattr__(name)

    def __delattr__(self, name):
        with VarsDBManager(self.db_fname) as data:
            if name in data:
                del data[name]
                return

        super().__delattr__(name)


################################################################################


class Workflow:
    """Workflow class representation

    It is:
    - a manager of blackdynamite sub-studies
    - a manager of event and firing actions
    """

    def __init__(self, filename):
        self.finished = False
        with open(filename) as f:
            self.config = yaml.load(f, Loader=yaml.SafeLoader)
            self.config_path = os.path.abspath(filename)
            self.directory = os.path.dirname(self.config_path)
        self.task_managers = dict(
            [(e, TaskManager(self, e)) for e in self.config["task_managers"]]
        )

        self.orchestrator_script = self.config["orchestrator"]
        self.orchestrator_function = None
        self.actions = {}

        conf_dir = os.path.join(self.directory, ".wf")
        os.makedirs(conf_dir, exist_ok=True)
        db_fname = os.path.join(conf_dir, "wf.db")
        self.vars = WorkflowVars(db_fname)

    def create(self, validated=None):
        import subprocess

        conf_dir = os.path.join(self.directory, ".wf")
        subprocess.run(f"rm -r {conf_dir}", shell=True)
        os.makedirs(conf_dir, exist_ok=True)
        db_fname = os.path.join(conf_dir, "wf.db")

        self.vars = WorkflowVars(db_fname)

        create_bd_studies(self, validated=validated)

    def start_launcher_daemons(self, args=None):
        # Select job management scheme (SLURM, PBS, bash, etc)
        # Default is bash

        config = self.config
        if args is None:
            import argparse

            args = argparse.Namespace()
            args.bd_study = "all"
        config.update(vars(args))

        clargs = ""
        if "generator" in self.config:
            generator = self.config["generator"]
            clargs += "--generator " + generator
            k = generator.replace("Coat", "_options")
            if k in self.config:
                clargs += " --" + k + " "
                clargs += " ".join(self.config[k])

        for name, task_manager in self.task_managers.items():
            detach = True
            if args.bd_study != "all" and args.bd_study != name:
                continue

            if args.bd_study == name and args.do_not_detach:
                detach = False

            cmd = f"canYouDigIt launch_daemon --start {clargs}"
            if detach:
                cmd += " -d"
            subprocess.call(
                cmd,
                cwd=task_manager.study_dir,
                shell=True,
            )

    def __getattr__(self, name):
        if name in self.task_managers:
            return self.task_managers[name]

        return super().__getattr__(name)

    def get_orchestrator_function(self):
        if self.orchestrator_function is not None:
            return self.orchestrator_function

        fname, func_name = self.orchestrator_script.split(".")
        file_path = os.path.join(self.directory, fname + ".py")
        module_name = "orchestrator"
        print(f"loading: {file_path}")
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
        self.orchestrator_function = getattr(module, func_name)
        return self.orchestrator_function

    def add_action(self, event_name, task="__all__", event=None, f=None):
        if (
            not self.is_run_filter(event)
            and not self.is_workflow_filter(event)
            and not isinstance(event, list)
            and not isinstance(event, str)
        ):
            raise RuntimeError(f"Unknown event description {event_name}")

        task_actions = self.actions.setdefault(task, {})
        task_actions[event_name] = event, f

    def add_error_handler(self, event="state = FAILED", f=no_action, **kwargs):
        def abort_orchestrator(*args, **kwargs):
            f(*args, **kwargs)
            print("ERROR: abort orchestrator")
            import sys

            sys.exit(-1)

        self.add_action("error_handler", event=event, f=abort_orchestrator, **kwargs)

    def is_run_filter(self, f):
        if not callable(f):
            return False

        import inspect

        sig = inspect.signature(f)
        params = sig.parameters
        names = [name for name, param in params.items()]
        if names == ["run", "job"]:
            return True
        return False

    def is_workflow_filter(self, f):
        if not callable(f):
            return False

        import inspect

        sig = inspect.signature(f)
        params = sig.parameters
        names = [name for name, param in params.items()]
        if names == ["workflow", "task_manager"]:
            return True
        return False

    @_transaction
    def fire_event(self, event_name, task_manager_name, f, trigger, params):
        if isinstance(trigger, bool):
            print(
                f"fire: event={event_name} "
                f"task_manager={task_manager_name} "
                f"val={trigger}"
            )
            f(workflow=self, **params)
        else:
            print(
                f"fire: event={event_name} "
                f"task_manager={task_manager_name} "
                f"val={[r.id for r, j in trigger]}"
            )
            f(runs=trigger, workflow=self, **params)

    def check_events(self, **params):
        for task_manager_name in self.actions:
            for event_name, (event, f) in self.actions[task_manager_name].items():
                if task_manager_name == "__all__":
                    task_managers = self.task_managers.items()
                    # will iterate over all tasks
                else:
                    task_managers = [
                        (task_manager_name, getattr(self, task_manager_name))
                    ]

                for task_manager_name, task_manager in task_managers:
                    if self.is_run_filter(event):
                        for run, job in task_manager.select([]):
                            is_fired = event(run, job)
                            if is_fired:
                                break
                    elif self.is_workflow_filter(event):
                        is_fired = event(self, task_manager=task_manager)
                    elif isinstance(event, list):  # must be a blackdynamite request
                        is_fired = task_manager.select(event)
                    elif isinstance(event, str):  # must be a blackdynamite request
                        event = [e.strip() for e in re.split(r",|\band\b", event)]
                        is_fired = task_manager.select(event)
                    else:
                        raise RuntimeError(f"Unknown event description {event_name}")

                    if is_fired:
                        self.fire_event(
                            event_name, task_manager_name, f, is_fired, params
                        )

    def execute(self, **params):
        """Orchestrator function that executes an action upon finishing a task"""

        func = self.get_orchestrator_function()
        func(self, **params)

        if not hasattr(self.vars, "stage"):
            self.vars.stage = "init"

        while not self.finished:
            self.check_events(**params)
            time.sleep(2)

    def get_runs(self, run_list):
        requests = {}
        for uri in run_list:
            task_manager_name, _id = uri.split(".")
            _id = int(_id)
            tm = self.__getattr__(task_manager_name)
            requests.setdefault(task_manager_name, []).append(tm.connect().runs[_id])
        return requests



#!/usr/bin/env python
# coding: utf-8

import math
import os
import pytest
import argparse
import subprocess

from motoko.scripts.create_studies import main as create_task_managers

################################################################
test_dir = os.path.dirname(__file__)
workflow_dir = os.path.join(test_dir, "workflow")
# sys.path.append(os.path.join(test_dir, "workflow"))
# from launch import main as main_daemon

################################################################


@pytest.fixture
def workflow():
    args = argparse.Namespace()
    args.directory = workflow_dir
    args.validated = True

    subprocess.call("motoko kill", shell=True, cwd=workflow_dir)
    workflow = create_task_managers(args)
    print("workflow task managers created")
    yield workflow
    subprocess.call("motoko kill", shell=True, cwd=workflow_dir)


def test_create_db(workflow):
    pass


def test_execute_pipeline(workflow):
    x = [2.1, 3.2]  # input

    print("Launch studies launcher daemons")
    workflow.start_launcher_daemons()
    print("Start the workflow")
    workflow.execute(inputs=x)
    print("commit results")
    subprocess.call("motoko info --verbose", shell=True, cwd=workflow_dir)

    val = (
        (0.342 * x[0] + 0.56 - 3.3) ** 2
        + (-2.31 * x[0] + 0.56 - 2.34) ** 2
        + (0.342 * x[1] + 0.56 - 0.148) ** 2
    )

    y1_ref = 0.342 * val
    y2_ref = 0.342 * math.sqrt(val)

    print("y1_ref =", y1_ref)
    print("y2_ref =", y2_ref)

    run_ids = [e for e in workflow.mult.base.runs.keys()]
    print("Runs", run_ids)
    assert max(run_ids) == 4

    y1 = workflow.mult.base.runs[3].y[0]

    print("y1 =", y1)

    assert math.fabs(y1 - y1_ref) < 1e-8



"""Scripts for creating and launching BlackDynamite studies"""

import os
import subprocess
import shutil
from BlackDynamite.scripts import createDB, bd_zeo_server
from BlackDynamite import bdparser

################################################################


def create_bd_study(task_manager):
    if not os.path.exists(task_manager.study_dir):
        raise RuntimeError(f"not existing study dir {task_manager.study_dir}")

    # Stop any existing BD server and delete files
    if os.path.exists(task_manager.study_dir + "/.bd"):
        subprocess.run(["canYouDigIt", "server", "stop"], cwd=task_manager.study_dir)
        # TODO: make a bd command to do the cleaning (to be working in remate also)
        subprocess.run(
            ["canYouDigIt", "reset", "--truerun", "--yes"], cwd=task_manager.study_dir
        )

    os.makedirs(task_manager.bd_dir, exist_ok=True)

    cwd = os.getcwd()
    os.chdir(task_manager.study_dir)
    # TODO: BD seems not to allow to create servers using tcp => fix this in BD
    # We should not have to stop and restart servers
    createDB.main(["--truerun", "--study", task_manager.study])
    bd_zeo_server.main(["--action", "stop"])
    bd_zeo_server.main(
        [
            "--action",
            "start",
            "--host",
            task_manager.host,
            "--study",
            task_manager.study,
        ]
    )
    os.chdir(cwd)


################################################################
def remote_launch(study, study_dir, study_files, host, ssh, local_dir):
    try:
        # Remove previous database
        cmd = f"mkdir -p {study_dir}; "
        cmd += f"cd {study_dir}; "
        cmd += "canYouDigIt server stop; "
        cmd += "rm -rf * .bd"
        ssh.cmd(cmd)
        ssh.copy(study_files, study_dir)
        # Initialize database
        cmd = f"cd {study_dir}; "
        cmd += "canYouDigIt init --truerun; "
        cmd += "canYouDigIt server stop; "
        cmd += f"canYouDigIt server start --host zeo://{host} --study {study}"
        ssh.cmd(cmd)
    except subprocess.CalledProcessError as e:
        print(e.stderr)

    # Assume that we want to access from the current work dir, create jobs/runs etc., we
    # need all files and a dummy .db folder in it
    if local_dir is not None:
        local_study_dir = local_dir + "/" + study
        os.makedirs(local_study_dir + "/.bd", exist_ok=True)
        for f in study_files:
            try:
                shutil.copy(f, local_study_dir)
            except shutil.SameFileError:
                pass


################################################################
def create_bd_studies(
    workflow,
    validated=None,
):
    if validated is None or not validated:
        validated = bdparser.validate_question(
            "Reset and drop content of Motoko database", {"yes": False}, False
        )
    if validated is False:
        return

    for name, task_manager in workflow.task_managers.items():
        create_bd_study(task_manager)


################################################################
class SSH:
    def __init__(self, host, user, port=None, profile=""):
        self.host = host
        self.user = user
        self.profile = profile

        self.remote = f"{user}@{host}"

        self.ssh_cmd = ["ssh"]
        self.scp_cmd = ["scp"]
        if port is not None:
            self.ssh_cmd += ["-p", f"{port}"]
            self.scp_cmd += ["-P", f"{port}"]

    def cmd(self, cmd):
        if self.profile == "":
            result = subprocess.run(self.ssh_cmd + [self.remote, cmd])
        else:
            result = subprocess.run(
                self.ssh_cmd + [self.remote, f"source {self.profile}; {cmd}"]
            )
        print(result.stdout)

    def copy(self, files, dest):
        result = subprocess.run(self.scp_cmd + files + [f"{self.remote}:{dest}"])
        print(result.stdout)



#!/usr/bin/env python3

from motoko.workflow import Workflow

command = "launcher"
command_help = "Spawn the launcher daemon"


def populate_arg_parser(parser):
    parser.add_argument(
        "--generator", type=str, help="Force a generator for job submission"
    )

    parser.add_argument("--log", "-l", action="store_true", help="Show logs")
    parser.add_argument(
        "--do_not_detach",
        action="store_true",
        help="Launcher start in front ground",
    )
    parser.add_argument(
        "--bd_study",
        type=str,
        default="all",
        help="The study launcher to start",
    )


def main(args):
    fullpath = "motoko.yaml"
    wf = Workflow(fullpath)

    if not args.log:
        wf.start_launcher_daemons(args)
        return

    import subprocess
    import os

    log_files = []
    for name, task_manager in wf.task_managers.items():
        _dir = task_manager.study_dir
        log_file = os.path.join(_dir, ".bd", "launch.log")
        log_files.append(log_file)

    log_files = " ".join(log_files)
    subprocess.call(f"tail -f {log_files}", shell=True)



#!/usr/bin/env python3

import os
from motoko.workflow import Workflow

command = "create"
command_help = "Create the sub studies of the workflow"


def populate_arg_parser(parser):
    parser.add_argument(
        "--validated",
        action="store_true",
        help="Answer yes to validation questions",
    )
    parser.add_argument(
        "directory",
        type=str,
        default="./",
        help="Directory where the description of the workflow is provided",
    )


def main(args):
    fullpath = os.path.join(args.directory, "motoko.yaml")

    wf = Workflow(fullpath)
    wf.create(args.validated)
    return wf



#!/usr/bin/env python3

import subprocess
import argparse
from motoko.workflow import Workflow
import sys

command = "info"
command_help = "Get info from sub-studies"


def populate_arg_parser(parser):
    parser.add_argument("--verbose", action="store_true", help="show verbose details")
    parser.add_argument(
        "--bd_study",
        "-bd",
        type=str,
        help="show details for a specific substudy",
        default="all",
    )
    parser.add_argument(
        "--byobu", action="store_true", help="show tiled window of the status"
    )
    parser.add_argument(
        "-",
        dest="bd_args",
        nargs=argparse.REMAINDER,
        help="Arguments to forward to canYouDigIt",
    )


def main(args):
    fullpath = "motoko.yaml"

    if args.byobu:
        import os

        session_name = os.path.abspath(fullpath)
        session_name = os.path.dirname(session_name)
        session_name = session_name.replace("/", "_")
        subprocess.run(f"byobu kill-session -t {session_name}", shell=True)
        subprocess.run(f"byobu new-session -d -s {session_name}", shell=True)
        subprocess.run(
            f'byobu send-keys -t {session_name} "watch motoko info" C-m', shell=True
        )
        subprocess.run(f"byobu split-window -h -t {session_name}", shell=True)
        subprocess.run(
            f'byobu send-keys -t {session_name}:0.1 "motoko launcher -l" C-m',
            shell=True,
        )
        subprocess.run(f"byobu split-window -v -t {session_name}", shell=True)
        subprocess.run(
            f'byobu send-keys -t {session_name}:0.2 "while \\$(test 1); do timeout 5s motoko info - runs exec --constraints state=START tail -f *.e* *.o*; done" C-m',
            shell=True,
        )
        subprocess.run(f"byobu attach -t {session_name}", shell=True)
        return

    wf = Workflow(fullpath)

    for name, task_manager in wf.task_managers.items():
        if args.bd_study != "all" and name != args.bd_study:
            continue
        print("*" * 30, file=sys.stderr)
        print(f"TaskManager: {name}", file=sys.stderr)
        print("*" * 30, file=sys.stderr)

        if hasattr(args, "bd_args"):
            cmd = "canYouDigIt " + " ".join(args.bd_args)
            subprocess.call(cmd.split(), cwd=task_manager.study_dir)
            continue

        if not args.verbose:
            subprocess.call("canYouDigIt info".split(), cwd=task_manager.study_dir)

        else:
            subprocess.call("canYouDigIt jobs info".split(), cwd=task_manager.study_dir)
            subprocess.call("canYouDigIt runs info".split(), cwd=task_manager.study_dir)
            subprocess.call(
                "canYouDigIt launch_daemon --status".split(), cwd=task_manager.study_dir
            )



#!/usr/bin/env python3

import os
import sys
import importlib.util
from BlackDynamite.scripts.launch_daemon import find_pids

from motoko.workflow import Workflow

command = "orchestrator"
command_help = "Start the workflow"


def populate_arg_parser(parser):
    subparsers = parser.add_subparsers(dest="what", required=True)
    parser_start = subparsers.add_parser("start", help="start orchestrator daemon")
    subparsers.add_parser("stop", help="start orchestrator daemon")
    parser_status = subparsers.add_parser("status", help="start orchestrator daemon")

    parser_status.add_argument(
        "--logs",
        "-l",
        action="store_true",
        help="Show logs of the orchestrator",
    )
    parser_status.add_argument(
        "--follow",
        "-f",
        action="store_true",
        help="Show logs and follow",
    )
    parser_start.add_argument(
        "--detach",
        "-d",
        action="store_true",
        help="For starting: run in detach/daemon mode with Zdaemon manager",
    )
    parser_start.add_argument(
        "--wait",
        type=str,
        default=5,
        help="Waiting time between state checks in seconds",
    )
    parser_start.add_argument(
        "--run_name",
        type=str,
        help="run_name to use for all produced jobs",
    )

    fullpath = "motoko.yaml"
    if not os.path.exists(fullpath):
        print(f"FATAL: not in a motoko directory (needs {fullpath})")
        sys.exit(-1)

    wf = Workflow(fullpath)

    fname, _ = wf.orchestrator_script.split(".")
    file_path = os.path.join(wf.directory, fname + ".py")
    module_name = "orchestrator"
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    parser_function = getattr(module, "populate_arg_parser")
    parser_function(parser_start)


def main(args):
    fullpath = "motoko.yaml"
    wf = Workflow(fullpath)
    wf_root_dir = wf.directory
    wf_conf_dir = os.path.join(wf_root_dir, ".wf")

    conf_fname = os.path.join(wf_conf_dir, "wf.conf")

    if not hasattr(args, "what"):
        import subprocess

        subprocess.call("motoko orchestrator --help", shell=True)
        sys.exit(-1)
    if args.what == "start" and not args.detach:
        params = vars(args)
        wf.execute(**params)
        sys.exit(0)

    elif args.what == "start" and args.detach:
        exclude = ["--detach", "-d"]
        argv = sys.argv[3:]
        clargs = " ".join([a for a in argv if a not in exclude])

        prog = f"motoko {command} start {clargs}"

        zdaemon_conf = f"""
<runner>
 program {prog}
 socket-name {wf_conf_dir}/wf.socket
 transcript {wf_conf_dir}/wf.log
 exit-codes 0
</runner>
"""

        os.makedirs(wf_conf_dir, exist_ok=True)
        with open(conf_fname, "w") as f:
            f.write(zdaemon_conf)

        os.system(f"zdaemon -C {conf_fname} start")
        os.system(f"zdaemon -C {conf_fname} status")

    elif args.what == "status":
        if not os.path.exists(conf_fname):
            print("Orchestrator daemon seeminlgy not run in detached mode")
            return
        print("ZDaemon status:")
        os.system(f"zdaemon -C {conf_fname} status")

        pids = find_pids(wf_root_dir)
        print(f"Running orchestrator pids: {pids}")
        if args.logs or args.follow:
            import subprocess

            cmd = f"tail {wf_conf_dir}/wf.log"
            if args.follow:
                cmd = f"tail {wf_conf_dir}/wf.log -f"
            else:
                cmd = f"cat {wf_conf_dir}/wf.log"
            subprocess.call(cmd, shell=True)

    elif args.what == "stop":
        if not os.path.exists(conf_fname):
            print(f"No Orchestrator daemon config {conf_fname}")
        else:
            os.system(f"zdaemon -C {conf_fname} stop")
        pids = find_pids(wf_root_dir)
        if pids:
            print(f"Killing: {pids}")
            pids = [str(e) for e in pids]
            os.system(f"kill -9 {' '.join(pids)}")
        else:
            print("No orphans found")



import subprocess
import json
import os


def start_daemon(script, inputs):
    cmd = ["python", script.__file__, json.dumps(inputs)]
    print("Execute: ", *cmd)
    return subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        cwd=os.getcwd(),
    )



#!/usr/bin/env python3

import subprocess

from motoko.workflow import Workflow

command = "kill"
command_help = "Kill all the running daemons"


def populate_arg_parser(parser):
    parser.add_argument("--verbose", action="store_true", help="show verbose details")
    parser.add_argument(
        "--bd_study",
        "-bd",
        type=str,
        default="all",
        help="The study launcher to kill",
    )
    parser.add_argument(
        "--reset",
        action="store_true",
        help="Reset completely the studies",
    )


def main(args):
    fullpath = "motoko.yaml"

    wf = Workflow(fullpath)

    for name, task_manager in wf.task_managers.items():
        if args.bd_study != "all" and args.bd_study != name:
            continue
        print(f"Kill daemons in study: {task_manager.study}")
        cmds = ["canYouDigIt launch_daemon --stop", "canYouDigIt server stop"]
        if args.reset:
            cmds += [f"rm -fr .bd BD-{name}-runs"]
        for cmd in cmds:
            if args.verbose:
                print(f"({task_manager.study}) {cmd}")
            subprocess.call(cmd, cwd=task_manager.study_dir, shell=True)
    if args.reset:
        subprocess.call("rm -rf .wf", shell=True)



import os
import math
import BlackDynamite as BD
import numpy as np

from motoko.workflow import Workflow


def doIt(run, job):
    mult_ids = job.mult_ids
    print(mult_ids)
    wf = Workflow(run.workflow)
    mult_runs = [wf.mult.connect().runs[id] for id in mult_ids]
    if len(mult_runs) != 2:
        raise ValueError("number of multi runs cannot be larger than 2")

    add_ids = []
    for r in mult_runs:
        add_ids.extend(r.dependencies)

    runs = wf.get_runs(add_ids)
    add_runs = runs["add"]
    print(add_runs)

    if len(add_runs) != 3:
        raise ValueError("number of add runs cannot be larger than 3")

    x = [0.0] * 3
    for i in range(3):
        x[i] = add_runs[i].getQuantity("scalar").iloc[0, 1]

    a = [3.3, 2.34, 0.148]
    print("AAAAA")
    print(a)
    print(x)
    l1_norm = (x[0] - a[0]) ** 2 + (x[1] - a[1]) ** 2 + (x[2] - a[2]) ** 2
    print(l1_norm)
    l2_norm = math.sqrt(l1_norm)
    print(l2_norm)
    run.pushQuantity(np.array([l1_norm, l2_norm]), 0, "vector")


################################################################

run, job = BD.getRunFromScript()
print(run)
print(job)
env_to_remove = []
for k in os.environ.keys():
    if "BLACKDYNAMITE_" in k:
        env_to_remove.append(k)

for k in env_to_remove:
    del os.environ[k]

run.start()
doIt(run, job)
run.finish()



import os
import BlackDynamite as BD

# from motoko.workflow import Workflow


def doIt(run, job):
    x = job.x

    n = 1
    if run.id == 1:
        n = 2

    a = [0.342, -2.31]

    y = [0.0] * n
    for i in range(n):
        y[i] = a[i] * x

    run.y = y


################################################################

run, job = BD.getRunFromScript()
print(run)
print(job)
env_to_remove = []
for k in os.environ.keys():
    if "BLACKDYNAMITE_" in k:
        env_to_remove.append(k)

for k in env_to_remove:
    del os.environ[k]

run.start()
doIt(run, job)
run.finish()



import os
import BlackDynamite as BD
import yaml
import copy


class TaskManager:
    def __init__(self, workflow, study):
        self.workflow = workflow
        self.study = study
        self.config = self.workflow.config["task_managers"][self.study]

        if self.config is None:
            self.config = {}
        if "host" in self.config:
            self.host = self.config["host"]
        else:
            self.host = "zeo://" + self.study_dir

        self._base = None
        self.selector = None

        self._default_job_space = {}
        self._default_run_params = {}
        if "job_space" in self.bd_config:
            self._default_job_space = self.bd_config["job_space"]
        if "run_params" in self.bd_config:
            self._default_run_params = self.bd_config["run_params"]

    def connect(self):
        return self.base

    @property
    def base(self):
        if self._base is not None:
            return self._base

        cwd = os.getcwd()
        os.chdir(self.study_dir)
        params = {"study": self.study, "host": self.host}
        self._base = BD.Base(**params)
        # print(self._base.schema)
        os.chdir(cwd)

        BD.singleton_base = self._base
        print(f"connected: to '{self.study}' => {self._base}")
        return self._base

    @property
    def bd_config(self):
        fname = os.path.join(self.study_dir, "bd.yaml")
        with open(fname) as f:
            config = yaml.load(f, Loader=yaml.SafeLoader)
        return config

    @property
    def study_dir(self):
        return os.path.join(self.workflow.directory, self.study)

    @property
    def bd_dir(self):
        return os.path.join(self.study_dir, ".bd")

    def _createRun(self, job, run_params):
        myrun = job.base.Run()
        # set the run parameters from the parsed entries
        job.base.prepare(myrun, "run_desc")
        myrun["run_name"] = "test"
        myrun["machine_name"] = "localhost"
        myrun["nproc"] = 1
        myrun["workflow"] = self.workflow.config_path
        myrun.entries.update(run_params)

        cwd = os.getcwd()
        os.chdir(self.study_dir)

        myrun.setExecFile("launch.sh")

        config_files = ["doIt.py"]
        # print(self.bd_config)
        if "config_files" in self.bd_config:
            for f in self.bd_config["config_files"]:
                config_files.append(f)
        myrun.addConfigFiles(config_files)
        os.chdir(cwd)

        _id = myrun.attachToJob(job)
        myrun = job.base.runs[_id]
        return myrun

    def _createJob(self, job_input):
        new_job = self.base.Job()
        for k, v in job_input.items():
            new_job[k] = v
        _id = self.base.insert(new_job)
        new_job = self.base.jobs[_id]
        return new_job

    def createTask(self, run_params=None, **kwargs):
        from BlackDynamite.base_zeo import BaseZEO

        BaseZEO.singleton_base = self.base

        if run_params is None:
            run_params = {}

        run_params_ = copy.deepcopy(self._default_run_params)
        for k, v in run_params.items():
            if isinstance(v, dict) and k in run_params_:
                run_params_[k].update(v)
            else:
                run_params_[k] = v

        job_space = self._expandJobSpace(**kwargs)
        return self._create_runs_and_jobs(job_space, run_params_)

    def _expandJobSpace(self, **kwargs):
        self.connect()
        desc_job = self.base.get_descriptor("job_desc")
        # print(desc_job)
        job_space = copy.deepcopy(self._default_job_space)
        job_space.update(kwargs)
        job_space = self.base._createParameterSpace(job_space, entries_desc=desc_job)
        return job_space

    def _create_runs_and_jobs(self, job_space, run_params):
        created = []
        for job_inputs in job_space:
            j = self._createJob(job_inputs)
            r = self._createRun(j, run_params)
            created.append((r, j))
        return created

    def select(self, constraints):
        if self.selector is None:
            self.selector = BD.RunSelector(self.base)
        selected_runs = self.selector.selectRuns(constraints, quiet=True)
        return selected_runs



import os
import BlackDynamite as BD


def doIt(run, job):
    x = job.x
    y = x + 0.56
    run.pushScalarQuantity(y, 0, "scalar")


################################################################

run, job = BD.getRunFromScript()
print(run)
print(job)
env_to_remove = []
for k in os.environ.keys():
    if "BLACKDYNAMITE_" in k:
        env_to_remove.append(k)

for k in env_to_remove:
    del os.environ[k]

run.start()
doIt(run, job)
run.finish()



import argparse

from motoko.scripts import create_studies, info, kill, launcher, orchestrator, clean

commands = [create_studies, info, kill, launcher, orchestrator, clean]


def main() -> None:
    """Entry point for the command line interface."""

    args = parse_args()

    keys = [e for e in args.__dict__.keys()]
    for k in keys:
        if args.__dict__[k] is None:
            del args.__dict__[k]

    try:
        for command in commands:
            if args.command == command.command:
                command.main(args)
                break
    except Exception as e:
        if args.verbose:
            raise e
        print(f"FATAL: {e}")


def parse_args():
    parser = argparse.ArgumentParser()
    parser.prog = "motoko"

    parser.add_argument("--verbose", action="store_true", help="verbose output")

    # Create subparsers for each command
    command_parsers = parser.add_subparsers(dest="command", help="command to run")
    command_parsers.required = True

    for command in commands:
        command_parser = command_parsers.add_parser(
            command.command, help=command.command_help
        )
        command.populate_arg_parser(command_parser)

    args = parser.parse_args()
    return args


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import subprocess

from motoko.workflow import Workflow

command = "clean"
command_help = "Clean all bd studies"


def populate_arg_parser(parser):
    parser.add_argument(
        "--delete", action="store_true", help="delete the runs (destructive)"
    )
    parser.add_argument(
        "--bd_study",
        "-bd",
        type=str,
        default="all",
        help="The study to clean",
    )
    parser.add_argument(
        "--constraints",
        type=str,
        help="The selection constraints",
    )


def main(args):
    fullpath = "motoko.yaml"
    wf = Workflow(fullpath)

    for name, task_manager in wf.task_managers.items():
        if args.bd_study != "all" and args.bd_study != name:
            continue

        print(f"clean study: {task_manager.study}")
        cmd = "canYouDigIt runs clean --truerun --yes"

        if hasattr(args, "constraints") and args.constraints:
            cmd += " --constraints {args.constraints}"
        if args.delete:
            cmd += " --delete"

        print(cmd)
        subprocess.call(cmd, cwd=task_manager.study_dir, shell=True)



from motoko.workflow import Workflow


def populate_arg_parser(parser):
    parser.add_argument(
        "--inputs",
        "-i",
        type=float,
        required=True,
        nargs=2,
    )


######################################################


def spawn_init_tasks(workflow, **kwargs):
    print("spawn_init_tasks")
    mult_manager = workflow.mult
    mult_manager.createTask(x=kwargs["inputs"], run_params=None)


######################################################


def spawn_add_tasks(runs=None, workflow=None, **kwargs):
    for run, job in runs:
        y = run.y
        created = workflow.add.createTask(x=y)
        run.state = "FORWARDED"
        run.dependencies = [f"add.{r.id}" for r, j in created]


######################################################


def spawn_norm_tasks(workflow=None, **kwargs):
    finished_mult_runs = workflow.mult.select(["state = FORWARDED"])
    mult_ids = [r.id for r, j in finished_mult_runs]
    for r, j in finished_mult_runs:
        for tm, add_runs in workflow.get_runs(r.dependencies).items():
            for add_run in add_runs:
                add_run.state = "FORWARDED"
    workflow.norm.createTask(mult_ids=mult_ids)


######################################################


def spawn_mult_tasks(runs=None, workflow=None, **kwargs):
    for r, j in runs:
        l1_norm, l2_norm = r.getQuantity("vector").iloc[0, 1:]
        print(l1_norm, l2_norm)
        created_mult_runs = workflow.mult.createTask(x=[l1_norm, l2_norm])
        r.dependencies = [f"mult.{r['id']}" for r, j in created_mult_runs]
        r.state = "FORWARDED"


######################################################


def dummy_task(workflow, **params):
    print("dummy_task: pass here")


######################################################


def finalize(workflow, **params):
    print("terminate: workflow successful")
    import sys

    sys.exit(0)


######################################################


def report_error(runs, **params):
    for r, j in runs:
        print(f"found error: {r.id}")
        print(f"run: {r}")
        import subprocess

        subprocess.run(
            f"cat {r.run_name}.*",
            shell=True,
            cwd=r.run_path,
        )


######################################################


def main(workflow, **params):
    workflow.add_error_handler(event="state = FAILED", f=report_error)

    workflow.add_action(
        "init",
        event=lambda workflow, task_manager: len(task_manager.select([])) < 2,
        task="mult",
        f=spawn_init_tasks,
    )

    workflow.add_action(
        "mult_finished",
        task="mult",
        event=["runs.id < 3", "state = FINISHED"],
        f=spawn_add_tasks,
    )

    def condition2(workflow, task_manager):
        if len(workflow.mult.select([])) != 2:
            return False
        if len(workflow.add.select([])) != 3:
            return False
        if workflow.mult.select(["state != FORWARDED"]):
            return False
        if workflow.add.select(["state != FINISHED"]):
            return False

        return True

    # for the others I show the most direct
    workflow.add_action(
        "need_launching_norm",
        event=condition2,
        f=spawn_norm_tasks,
    )

    workflow.add_action(
        "norm_run_finished",
        task="norm",
        event="state = FINISHED",
        f=spawn_mult_tasks,
    )

    def condition3(workflow, task_manager):
        if len(workflow.mult.select([])) != 4:
            return False
        if len(workflow.add.select([])) != 3:
            return False
        if len(workflow.norm.select([])) != 1:
            return False
        if workflow.mult.select(["state != FORWARDED", "state != FINISHED"]):
            return False
        if workflow.add.select(["state != FORWARDED"]):
            return False
        if workflow.norm.select(["state != FORWARDED"]):
            return False

        return True

    workflow.add_action(
        "finish",
        event=condition3,
        f=finalize,
    )


if __name__ == "__main__":
    workflow = Workflow("motoko.yaml")
    main(workflow)
